/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "fh2805config.h"

#include <sys/ioctl.h>
#include <sys/types.h>
#include <pthread.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <portLib.h>
#include <taskLib.h>

#include "fh2805/fh2805.h"
#include "fh2805Cmds.h"
#include "serial.h"

/***
 *** Main interface for the FaulHaber 2805 motor controller
 ***/

static void fh2805ReadTask(fh2805dev_t *);

/*----------------------------------------------------------------------*/

/**
 ** Initialize the library
 **
 ** Allocates the device descriptor, opens the serial line and spawns
 ** the task that reads from this line.
 ** As a test, the 'get version' command is sent and its reply is displayed
 **
 ** Returns: a pointer to the device descriptor, or NULL in case of a error.
 **/
fh2805dev_t *
fh2805Init(char *name, int baud)
{
	fh2805dev_t *dev;

	dev = (fh2805dev_t *)malloc(sizeof(fh2805dev_t));
	if (dev == NULL) {
		return NULL;
	}
	dev->buf = malloc(FH2805_BUF_SIZE);
	if (dev->buf == NULL) {
		goto fail;
	}
	dev->mutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));
	if (dev->mutex == NULL) {
		goto fail;
	}
	pthread_mutex_init(dev->mutex, NULL);
	dev->decodeFunc = NULL;
	dev->asyncAnswer = '\0';
	dev->fd = openSerial(name, baud);
	if (dev->fd < 0) {
		goto fail;
	}
	dev->readTask = taskSpawn(FH2805_READ_TASK_NAME, 
				  FH2805_READ_TASK_PRIORITY, 
				  VX_FP_TASK, 
				  FH2805_READ_TASK_STACK_SIZE,
				  (FUNCPTR)fh2805ReadTask,
				  dev, NULL);
	if (dev->readTask == ERROR) {
		close(dev->fd);
		goto fail;
	}
	fh2805Cmd(dev, GET_VERSION);
	while (fh2805Busy(dev))
		;
	printf("%s\n", dev->version);

	return dev;
fail:
	if (dev->mutex != NULL) free(dev->mutex);
	if (dev->buf != NULL) free(dev->buf);
	free(dev);
	return NULL;
} /* fh2805Init */

/*----------------------------------------------------------------------*/

/**
 ** Close the library
 **
 ** kill the reading task, frees allocated resources and 
 ** close the serial line
 **
 ** returns: OK (or ERROR)
 **/
int
fh2805End(fh2805dev_t *dev)
{
	taskDelete(dev->readTask);
	pthread_mutex_destroy(dev->mutex);
	free(dev->mutex);
	close(dev->fd);
	free(dev->buf);
	free(dev);
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 **  Send a command and register callback for the reply
 **
 ** Returns: OK or ERROR
 **/
int
fh2805Cmd(fh2805dev_t *dev, FH2805_CMD cmd, ...)
{
	va_list ap;
	FH2805_CMD_TAB *c;
	char buf[256];
	size_t l;

	c = fh2805FindCmd(cmd);
	if (c == NULL)
		return ERROR;

	pthread_mutex_lock(dev->mutex);
	if (c->asyncAnswer != '\0') {
		dev->asyncAnswer = c->asyncAnswer;
		dev->decodeFunc = NULL;
	} else {
		dev->decodeFunc = c->decodeFunc;
	}
	va_start(ap, cmd);
	vsnprintf(buf, sizeof(buf), c->format, ap);
	va_end(ap);
	l = strlen(buf);
	if (l == sizeof(buf)) { 
		pthread_mutex_unlock(dev->mutex);
		return ERROR;
	}
	buf[l++] = '\r';
	if (write(dev->fd, buf, l) != l) {
		pthread_mutex_unlock(dev->mutex);
		return ERROR;
	}
	pthread_mutex_unlock(dev->mutex);
	return OK;
} /* fh2805Cmd */


/*----------------------------------------------------------------------*/

/**
 ** Send ASCII commands directly 
 **
 ** Used for debugging only
 **
 ** Returns: OK or ERROR
 **/
int 
fh2805SendCmd(fh2805dev_t *dev, char *fmt, ...)
{
	va_list ap;
	char buf[256];
	size_t l;

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
	l = strlen(buf);
	if (l == sizeof(buf)) {
		return ERROR;
	} else {
		buf[l] = '\r';
	} 
	if (writeSerial(dev->fd, buf, l+1) != l+1) {
		return ERROR;
	}
	return OK;
} /* fh2805SendCmd */
	
/*----------------------------------------------------------------------*/

/** 
 ** Checks if the controller is busy
 **
 ** Returns: TRUE if busy, FALSE if available
 **/
int 
fh2805Busy(fh2805dev_t *dev)
{
	int ret;
	pthread_mutex_lock(dev->mutex);
	ret = (dev->decodeFunc != NULL || dev->asyncAnswer != 0);
	pthread_mutex_unlock(dev->mutex);
	return ret;
}

/*----------------------------------------------------------------------*/

/**
 ** Code for the task that reads the serial line and decodes messages
 **
 ** Returns: never
 **/
static void
fh2805ReadTask(fh2805dev_t *dev)
{
	int fd, nread;

	/* Get a local copy of the file descriptor */
	pthread_mutex_lock(dev->mutex);
	fd = dev->fd;
	pthread_mutex_unlock(dev->mutex);

	do {
		/* Wait for data on the line */
		nread = readSerial(fd, dev->buf, FH2805_BUF_SIZE);
#ifdef FH2805_DEBUG
		fprintf(stderr, "%s: read %d\n", __func__, nread);
#endif
		if (nread > 0) {
			/* NUL-terminate the string */
			if (nread < FH2805_BUF_SIZE) 
				dev->buf[nread] = '\0';
			else {
				dev->buf[FH2805_BUF_SIZE - 1] = '\0';
				fprintf(stderr, "%s: buffer overrun\n", 
				    __func__);
			}
			/* Decode the result */
			pthread_mutex_lock(dev->mutex);
			if (dev->decodeFunc != NULL) {
				if (dev->decodeFunc(dev) != OK) {
					fprintf(stderr, 
					    "%s: decodeFunc returns ERROR\n",
					    __func__);
				}
				/* Mark the result as decoded */
				dev->decodeFunc = NULL;
			} else {
				fprintf(stderr, "%s: no decodeFunc, "
				    "received %s\n",
				    __func__, dev->buf);
				fh2805CheckAsyncAnswer(dev);
			}
			pthread_mutex_unlock(dev->mutex);
		} else {
			/* Nothing was read, so this is an error */
			/* XXX it should be handled! */
			fprintf(stderr, "%s: read returns %d\n",
			    __func__, nread);
		}
	} while (1);
} /* fh2805ReadTask */
