/*	$LAAS$ */
/*
 * Copyright (c) 2003-2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "serial.h"

#ifdef __sun
static void cfmakeraw(struct termios *);
static int cfsetspeed(struct termios *, speed_t);
#endif

/**
 ** numerical baud rate to termios(7) constants conversion table
 **/
static struct speed {
        int speed;
        speed_t termioSpeed;
} speeds[] = {
        { 0, B0 },
        { 50, B50 },
        { 75, B75 },
        { 110, B110 },
        { 300, B300 },
        { 600, B600 },
        { 1200, B1200 },
        { 2400, B2400 },
        { 4800, B4800 },
        { 9600, B9600 },
        { 19200, B19200 },
        { 38400, B38400 },
#ifdef B57600
        { 57600, B57600 },
#endif
#ifdef B115200
        { 115200, B115200 }
#endif
};

#define SPEED_SIZE (sizeof(speeds)/sizeof(struct speed))

/**
 ** Get the termios(7) constant for a numerical baud rate value
 **/
static speed_t
get_termios_speed(int speed) 
{
        int i;

        for (i = 0; i < SPEED_SIZE; i++) {
                if (speed == speeds[i].speed) {
                        return speeds[i].termioSpeed;
                }
        } /* for */
        return -1;
}


/**
 ** Open a serial line for reading and writing, 
 ** set it to raw mode at given baud rate
 **
 ** returns: the open file descriptor
 **          -1 on error
 **/
int 
openSerial(const char *device, int baud)
{
	int fd, flags;
	struct termios t;

	fd = open(device, O_RDWR|O_NONBLOCK);
	if (fd < 0) {
		fprintf(stderr, "%s:%d: %s: %s\n", __FILE__, __LINE__,
		    device, strerror(errno));
		return -1;
	}
	tcgetattr(fd, &t);
	cfmakeraw(&t);

	cfsetspeed(&t, get_termios_speed(baud));
	if (tcsetattr(fd, TCSAFLUSH, &t) < 0) {
		fprintf(stderr, "%s:%d: tcsetattr: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}
	flags = fcntl(fd, F_GETFL, 0);
	if (flags < 0) {
		fprintf(stderr, "%s:%d: F_GETFL: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}
	flags &= ~O_NONBLOCK;
	if (fcntl(fd, F_SETFL, flags) < 0) {
		fprintf(stderr, "%s:%d: F_SETFL: %s\n", __FILE__, __LINE__,
		    strerror(errno));
		close (fd);
		return -1;
	}
	
	return fd;
}

/**
 ** Blocking read one character from a serial line
 **
 ** Returns: the character read 
 **          -1 on error
 **/
int
serialGetChar(int fd)
{
	char buf;
	int n;
	
	errno = 0;
	do {
		n = read(fd, &buf, 1);
	} while (n < 0 && errno == EINTR);

	if (n < 0) {
		fprintf(stderr, "%s:%d: read: %s\n",
		    __FILE__, __LINE__, strerror(errno));
		return -1;
	}
	return (int)buf;
}

/**
 ** Reads up to 'len' bytes of data from a serial line
 **
 ** Returns: the number of bytes actually read
 **          -1 on error
 **/
int
readSerial(int fd, char *buf, size_t len)
{
#if 1
	int i = 0, n;
	char c;
	
	do {
		errno = 0;
		do {
			n = read(fd, &c, 1);
		} while (n < 0 && errno == EINTR && i < len);
		
		buf[i++] = c;
		if (c == '\n' || i == len) 
			break;
	} while (1);
	return i;
#else
	return read(fd, buf, len);
#endif
}

/**
 ** Write a buffer of 'len' bytes to the serial line
 **
 ** Returns: the actual number of written bytes
 **           -1 on error
 **/
int 
writeSerial(int fd, char *buf, size_t len)
{
	/* for now */
	return write(fd, buf, len);
}

#ifdef __sun
/**
 ** These functions are missing in Solaris 
 **/
static void
cfmakeraw(struct termios *t)
{
        t->c_iflag &= ~(IMAXBEL|IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
        t->c_oflag &= ~OPOST;
        t->c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
        t->c_cflag &= ~(CSIZE|PARENB);
        t->c_cflag |= CS8;
        /* XXX set MIN/TIME */
}

static int
cfsetspeed(struct termios *t, speed_t speed)
{
	cfsetispeed(t, speed);
	cfsetospeed(t, speed);
        return (0);
}
#endif
